package de.fraunhofer.iem.dataprovider.kpi.metrics

import de.fraunhofer.iem.dataprovider.kpi.metrics.repositoryDetails.service.RepositoryDetailsService
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.toolRun.service.ToolRunService
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.async
import kotlinx.coroutines.launch
import org.springframework.stereotype.Service

@Service
class MetricsService(
    private val repositoryService: RepositoryService,
    private val toolRunService: ToolRunService,
    private val repositoryDetailsService: RepositoryDetailsService,
    private val kpiService: KPIService
) {

    private val defaultScope = CoroutineScope(Dispatchers.Default)
    val ioScope = CoroutineScope(Dispatchers.IO)

    suspend fun handleRepositoryChange(repoId: Long) = defaultScope.launch {

        // TODO: we need to create the tools in the db and link them to the tool run.
        // This should probably be done on the initial server startup, based on some
        // config properties, as the tools don't change dynamically.

        /**
         * Create initial db models.
         * If this fails we want the whole method to fail as we can't continue
         * without the db models.
         *
         */
        val toolRun = async(ioScope.coroutineContext) {
            val repoDto = repositoryService.getRepositoryInfo(repoId)
            val repo = repositoryService.getOrCreate(repoDto)
            toolRunService.createToolRunForRepository(repo)
        }.await()

        /**
         * For all tools, first query the tool results API.
         * If the tool has results, add the tool to the tool run
         * and then calculate the KPI based on the results, link
         * the KPI with the repository and store everything in the db.
         * The function returns the UUID of the KPI in the db.
         * If we can calculate multiple KPIs from one tool we can return
         * a list of UUIDs.
         *
         * val toolKpi = async {
         *  val toolResult = queryToolApi(repoId)
         *  if (!toolResult.isEmpty) {
         *      toolRun.addTool(tool)
         *      val kpi = calculateKpi(toolResults)
         *      kpi.addRepository(toolRun.repo)
         *      kpi.store()
         *      kpi.id // this is the return of the async function
         *  }
         * }
         */


        val repositoryDetailsKpi = async {
            val repoDetailsDto = repositoryDetailsService.getRepositoryDetails(repoId)
            repositoryDetailsService.calculateRepositoryDetailsKpis(repoDetailsDto)
        }

        /**
         * Calculate KPI tree by awaiting all tool KPIs
         * joinAll(toolKpi)
         * calculateKpiTree(toolKpi, ...)
         */
        kpiService.calculateAndSaveKpiTree(toolRun.repository!!, repositoryDetailsKpi.await())
    }

}